#include "liblegfct.h"

#define eps 1e-15
#define MIN(a,b) ((a)>(b)?(b):(a))	       
#define SIGN(x)  ((x)>(0)?(1.0):(-1.0))

/* checkEllipse 
 * Check if a complex point z is inside of an ellipse
 * with minor axix b = min(1, 3 / (p+1)^1.2
 */
int checkEllipse(double u2, double v2, int p){
    double b2,b,a2;
    b = MIN(1,3/pow(p+1.0,1.2));   
    b2 = b*b;  
    a2 = 1.0+b2;    
    return (u2/a2+v2/b2 < 1);
}


/* forward
 * Forward recurrence relation  
 *  for j=start:p-1
 *     V(j+1)=((2*j+a)*z*V(j)-(j+c)*V(j-1))/(j+d);
 *   end
 */
void  forward(int start, double a, double b, double c,  \
              double u, double v, int p, double *Vre, double *Vim){
  int j;
  double tmp1, tmp2, tmp3;
  for (j=start; j<=p-1; j++){
    tmp1 = (2.0*j+a);
    tmp2 = j+b;
    tmp3 = j+c;
    Vre[j+1] = (tmp1*(u*Vre[j]-v*Vim[j])-tmp2*Vre[j-1])/tmp3;
    Vim[j+1] = (tmp1*(u*Vim[j]+v*Vre[j])-tmp2*Vim[j-1])/tmp3;
  }   
}


/* backward
 * Gautschi's continued fraction algorithm
 */
void backward (int n0, double a, double b, \
                      double u, double v, int p, double *Vre, double *Vim){
  double x, y;
  double tmp1, tmp2,tmp3, absz;
  int n, k, k1, k2; 
  int nu;
  fcomplex z2, tmp;
  absz = sqrt(u*u+v*v);
  /* Assign nu = p + ceil( 8 / log10|z+sqrt(z^2-1)| )*/
  z2 = Complex(fabs(u),fabs(v));
  tmp = Csqrt(Cadd(Complex(-1.,0.), Cmul(z2,z2)));	      
  nu = (int) (p+ceil(8. / (log10(Cabs(Cadd(z2, tmp)))))); 
  nu = ceil(1.1*nu);
  
  /* Calculate vector r */
  x = 1.0; y = 0.0;
  for (k=nu; k>=p-n0; k--){   
    tmp1 = u-x;
    tmp2 = v-y;
    tmp3 = (k*k-a)/((4.*k*k-1.)*(tmp1*tmp1 + tmp2*tmp2));
    x =  tmp3*tmp1;
    y = -tmp3*tmp2;  
  }
  Vre[p] = x; Vim[p] = y;  /* V[p] = r[p-n0-2] */
  for (n=p; n>=3+2*n0; n--){   
    k = n-1-n0;
    tmp1 = u-Vre[n];
    tmp2 = v-Vim[n];
    tmp3 = (k*k-a)/((4.*k*k-1.)*(tmp1*tmp1 + tmp2*tmp2));
    Vre[n-1] =  tmp3*tmp1;
    Vim[n-1] = -tmp3*tmp2;    
  }  
  for (n=2+n0; n<p+1-n0; n++){  /* Calculate values of V */
    tmp1 = (2.*n-1.)/(n+b);
    tmp2 = Vre[n+n0];
    Vre[n+n0] = tmp1*(tmp2*Vre[n-1+n0] - Vim[n+n0]*Vim[n-1+n0]);
    Vim[n+n0] = tmp1*(tmp2*Vim[n-1+n0] + Vim[n+n0]*Vre[n-1+n0]);
  }
}


/* horner
 * Evaluate a complex polynomial using the Horner scheme 
 */
void horner(double zre, double zim, double* coeff, 
            int n, double* pre, double* pim)
{
  int  i;
  pre[0] = 0.0;
  pim[0] = 0.0;
  if (zim==0.0){/* real evaluation */
    pre[0] = coeff[n-1];
    for(i = n-2; i >=0; i--){
	    pre[0] = zre*pre[0] + coeff[i];
	  } 
  }else{ /* complex evaluation*/
		double retmp;
		pre[0] = coeff[n-1];
		pim[0] = 0.;
		for(i=n-2; i>=0; i--){
			retmp  = pre[0]*zre - pim[0]*zim + coeff[i];
			pim[0] = pim[0]*zre + pre[0]*zim;
			pre[0] = retmp;
		}
  }
}

/* Taylor expansion of log(1+x) */
/*
double taylor_log1px(double x)
{
    double xtmp = x;
    double x2 = x*x;
    double logtmp = 0;
    int k;
    for(k=1; k<100; k+=2){
       logtmp += xtmp/(k*(k+1.0)) * (k+1.0-k*x);   
       xtmp = xtmp*x2;
    }
    return logtmp;
}

double get_logtmp(double u, double v, 
                  double u2, double v2, 
                  double um1, double up1)
{ 
    double logtmp;
    if(fabs(u)<0.2*fabs(v)+0.05){
      logtmp = 0.5*taylor_log1px(4*u/(v2+um1*um1));
    } else {
      double b1,b2;
      b1 = up1*up1+v2;
      b2 = um1*um1+v2;
      logtmp = log(b1/b2)*0.5;    
    }
    return logtmp;
}
*/

/* phiinv
 * calculate w = phi^{-1}(z) = z + sign(u) * sqrt(z^2-1); 
 */
void phiinv(double u, double v, double *x, double *y)
{
  fcomplex z = Complex(u,v), w;
	w = Csqrt(Csub(Cmul(z,z),Complex(1,0)));
  if(u>0)	      w = Cadd(z,w);
  else if(u<0)  w = Csub(z,w);
	else{
		if(v>0) w.i = v+sqrt(v*v+1), w.r = 0.0;
		else    w.i = v-sqrt(v*v+1), w.r = 0.0;
	}

	x[0] = w.r;
	y[0] = w.i;
}


/* get_logtmp 
 * calculate Re(log(z+1)-log(z-1)) 
 */
double get_logtmp(double u, double v, double u2, double v2, 
                  double um1, double up1)
{ 
    double logtmp;
    double b1,b2;
    b1 = up1*up1+v2;
    b2 = um1*um1+v2;
    logtmp = log(b1/b2)*0.5;    
    return logtmp;
}

/* get_argtmp
 * calculate Im(log(z+1)-log(z-1))
 */
double get_argtmp(double u, double v, 
                  double u2, double v2, 
                  double um1, double up1)
{
  double absz = sqrt(u2+v2);
  double abszp1 = sqrt(up1*up1 + v2);
  double abszm1 = sqrt(um1*um1 + v2);
	double argtmp;
  if (absz==1){
    argtmp = -SIGN(v)*M_PI/2.0;
  } else {
	  double r,phi,tk,rk;
		int k;
    if(abszp1 < 0.3){
 		  phi = atan2(v,up1); /* arg(z+1) */
		  r = abszp1;         /* |z+1|    */
		  tk = 2., rk = r;
		  argtmp = phi - ((v>0)?1:-1)*M_PI;		
		  for(k=1; k<22; k++){
			  argtmp += rk/(k*tk)*sin(k*phi);
			  tk*=2.;	rk*=r;
		  }
    }else if(abszm1<0.3){
      phi = atan2(v,um1);
		  r = abszm1;
		  tk = 2., rk = r;
		  argtmp = -phi;		
		  for(k=1; k<22; k++){
			  argtmp -= rk/(k*tk)*sin(k*(phi+M_PI));
			  tk*=2.;	rk*=r;
		  }       
    }else{
      argtmp = atan(-2.0*v/(absz*absz-1.0));
      if(absz<1) argtmp+= SIGN((um1)/v)*M_PI;
    }
  }      
  return argtmp;  
}

/* get_taylor_deg
 * returns the upper limit for the Laurent series 
 */
int get_taylor_degk(double absz2){
	int nu;
	if      (absz2>4096) nu =  7;
	else if (absz2> 256) nu =  9;
	else if (absz2>  64) nu = 12;
	else if (absz2>  16) nu = 18;
	else if (absz2>   4) nu = 33; 
	else if (absz2>2.25) nu = 56; 
	return nu;
}

int get_taylor_degik(double absz2){
	int nu;
	if      (absz2>4096) nu =  6;
	else if (absz2> 256) nu =  8;
	else if (absz2>  64) nu = 11;
	else if (absz2>  16) nu = 16;
	else if (absz2>   4) nu = 30; 
	else if (absz2>2.25) nu = 50; /* r1 braucht hier 50 !!! */
	return nu;
}

int get_taylor_degik3(double absz2){
	int nu;
	if      (absz2>4096) nu =  6;
	else if (absz2> 256) nu =  8;
	else if (absz2>  64) nu = 10;
	else if (absz2>  16) nu = 14; 
	else if (absz2>   4) nu = 24;
	else if (absz2>2.25) nu = 38;
	return nu;
}

/* qtm1
 *
 * INPUT:	u,v     (point of evaluation z = u+iv)
 *          p       (max. polynomial degree)
 *          fac     (factor to multiply)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
        Vim[2] = 0.0;            

 *         / 1
 *         |
 *   fac * |  P_k(t) * log(z-t) dt;   z = u+iv, k = 0, ..., p
 *         |
 *         / -1
 *
 *   where P_k(t) are the Legendre polynomials
 *
 */
void 
qtm1(double u, double v, int const p, double const fac, 
     double *Vre, double *Vim)
{ 

  double u2, v2, tmp, absz2, x, y, absw2; 
  u2 = u*u;
  v2 = v*v; 
  absz2 = u2+v2;
  
	phiinv(u,v,&x,&y);
	absw2 = x*x+y*y;

  /* Nearfield: |w| < 1.5 */
  if(absw2<2.25){
    double b1, b2, argtmp, argtmp1, logtmp, up1, um1;
    um1 = u-1.0;
  	up1 = u+1.0;
  	b1 = up1*up1+v2;
  	b2 = um1*um1+v2;
  	logtmp = log(b1/b2)*0.5;    
  	/*logtmp = get_logtmp(u, v, u2, v2, um1, up1); */

    if ( fabs(v) > eps ){ /* case: z is complex*/
      argtmp = get_argtmp(u, v, u2, v2, um1, up1);
			argtmp1 = atan2(v,up1)+atan2(v,um1);
			
			Vre[0] =fac*( u*logtmp - v*argtmp + 0.5*log(b1*b2) - 2.);
      Vim[0] =fac*( v*logtmp + u*argtmp + argtmp1); 
      if(p>0){
        tmp = 0.5*(u2-v2-1);
        Vre[1] =fac*( logtmp*tmp-u - u*v*argtmp );
				Vim[1] =fac*( u*v*logtmp + tmp*argtmp-v );
      }
    
		}else if(fabs(fabs(u)-1)>eps){ /* case: z is real, z != +/- 1 */
      Vre[0] =fac*(log(fabs(up1)*fabs(um1))+u*logtmp -2.0);
      Vim[0] = 0.0;
      if(p>0){
       Vre[1] = fac*(0.5*(u2-1)*logtmp-u);
       Vim[1] = 0.0;
     }
    } else {   /* case: z = +/- 1 */	
      Vre[0] = fac*(2.*log(2.)-2.); 
      Vim[0] = 0.0;
      if(p>0){
			  if (u>0) Vre[1] = -fac, Vim[1] = 0.0;
			  else     Vre[1] =  fac, Vim[1] = 0.0;
      }
    }

  /* Farfield |w| >=1.5 --> Laurent series */
  }else{
  	double coeffs0[40], coeffs1[40];
  	int k,nu;

	  nu = get_taylor_degik3(absw2);  
    coeffs0[0] = 0;
    for (k=1; k<nu; k++)  
		  coeffs0[k] = 2.0/(k*(2.*k-1.)*(2.*k+1.));
   
	  if(p>0)
	    for (k=0; k<nu; k++) 
		    coeffs1[k] = 4.0/((2.*k-1.)*(2.*k+1.)*(2.*k+3.));

	  if ( fabs(v) > eps ){ /* case: z is complex */
      tmp = absw2*absw2;
			/* Horner scheme, x = 1/z^2 */
      horner((x+y)*(x-y)/tmp,-2*x*y/tmp,coeffs0,nu,Vre,Vim);
      Vre[0] = fac*(-2*log(2)+Vre[0]+log(absw2));
      Vim[0] = fac*(Vim[0]+2*atan2(y,x));        
    
      if(p>0){          
        fcomplex ctmp;
        /* Horner scheme, x = 1/z^2 */
        horner((x+y)*(x-y)/tmp,-2*x*y/tmp,coeffs1,nu,&ctmp.r,&ctmp.i);
        ctmp = Cmul(ctmp,Complex(x/absw2,-y/absw2));
        Vre[1] = fac*ctmp.r;
        Vim[1] = fac*ctmp.i;                
      }
    
    } else { /* case: z is real, z != +/- 1 */
      tmp = 1.0/x;
      horner(tmp*tmp,0.0,coeffs0,nu,Vre,Vim);
      Vre[0] = fac*(-2*log(2)+Vre[0]+log(absw2));         
      if(p>0){
        horner(tmp*tmp,0.0,coeffs1,nu,Vre+1,Vim+1);
        Vre[1] =  fac*tmp*Vre[1];     
      }
      for(k=0; k<p; k++) 
			  Vim[k]=0.0; 
		}
  }

  /*Calculate V[2]..V[p]*/   
  if(p>1){ 
	  Vre[2] = u*Vre[1]-v*Vim[1]+ (fac*2.0)/3.0;
	  Vim[2] = u*Vim[1]+v*Vre[1]; 

    /* Inside of ellipse, calculation via forward recurrence */
    if (checkEllipse(u2,v2,p)){  /*inside */        
      forward(2, 1.0, -1.0, 2.0, u, v, p, Vre, Vim); /* V[3] ... V[p] */ 
    } else { /*outside*/
      backward (0, 1.0, 1.0, u, v, p, Vre, Vim);     /* V[2] ... V[p] */  
    }
  } 
}


/* qt0
 *
 * INPUT:	u,v     (point of evaluation z - u+iv)
 *          p       (max. polynomial degree)
 *          fac     (factor to multiply)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *        / 1
 *        |
 *  fac * |  P_k(t) / (z-t) dt;   z = u+iv, k = 0, ..., p
 *        |
 *        / -1
 *
 *   where P_k(t) are the Legendre polynomials.
 *
 */
void qt0(double u, double v, int const p, double const fac, 
         double *Vre, double *Vim){
  int j;
  double u2, v2, absz2, tmp,  x, y, absw2;  
  u2 = u*u;
  v2 = v*v;
  absz2 = u2+v2;

	phiinv(u,v,&x,&y);
	absw2 = x*x+y*y;

	if(absw2 < 2.25){
    double logtmp, argtmp, um1, up1;
    um1 = u-1.0;
    up1 = u+1.0;
 
    if ( fabs(v) > eps ){ /* z is complex */
      logtmp = get_logtmp(u, v, u2, v2, um1, up1); 
      argtmp = get_argtmp(u, v, u2, v2, um1, up1); 
      Vre[0] = fac*logtmp;
      Vim[0] = fac*argtmp;
      if(p>0){
        Vre[1] = fac*(u*logtmp - v*argtmp - 2.0);
        Vim[1] = fac*(u*argtmp + v*logtmp);
      }  
    } else if(fabs(fabs(u)-1)>eps){ /* z is real and not +-1*/  
      logtmp = log(fabs((up1)/(um1)));
      Vre[0] = fac*logtmp;
      Vim[0] = 0.0;
      if(p>0){
        Vre[1] = fac*(u*logtmp - 2.0);
        Vim[1] = 0.0; 
      }
    } else { /* z = +/- 1 (Singularity --> Set values to 0)*/  
      for(j=0; j<=p; j++){
        Vre[j] = 0.0;
        Vim[j] = 0.0;
      }
      return;
    }
  
  }else {
    double coeffs0[52], coeffs1[52]; 
  	int k,nu;
    nu = get_taylor_degik(absw2);  
    
		for (k=0; k<nu+1; k++){  
      coeffs0[k] = 4.0/(2.*k+1.);
			coeffs1[k] = 8.0*k/((2.*k-1.)*(2.*k+1.));
		}
		
    if ( fabs(v) > eps ){ /* z is complex */
      fcomplex ctmp;
			tmp = absw2*absw2;
      /* Horner scheme,  1/w^2 */
      horner((x+y)*(x-y)/tmp,-2*x*y/tmp,coeffs0,nu,&ctmp.r,&ctmp.i);
      ctmp = Cmul(ctmp,Complex(x/absw2,-y/absw2));
      Vre[0] = fac*ctmp.r;
      Vim[0] = fac*ctmp.i;             
      if(p>0){
        /* Horner scheme, 1/w^2 */
        horner((x+y)*(x-y)/tmp,-2*x*y/tmp,coeffs1,nu,Vre+1,Vim+1);
        Vre[1] = fac*Vre[1];
        Vim[1] = fac*Vim[1]; 
      }  
    } else{ /* z is real and not +-1*/
		  tmp = 1./x;
      horner(tmp*tmp,0,coeffs0,nu,Vre,Vim);
      Vre[0] = fac*tmp*Vre[0];
      Vim[0] = 0.0;
      if(p>0){
        horner(tmp*tmp,0,coeffs1,nu,Vre+1,Vim+1);
        Vre[1] = fac*Vre[1];   
        Vim[1] = 0.0; 
      }
    }       
  }

  /* Calculate V[2]..V[p] */
  if(p>1){         
    /* Inside ellipse, calculation via forward recurrene */
    if (checkEllipse(u2,v2,p)) { /*inside*/
      forward(1, 1.0, 0.0, 1.0, u, v, p, Vre, Vim); 
    } else { /*outside*/
      backward(0, 0.0, 0.0, u, v, p, Vre, Vim);    
    }
  }  
}


/* qt1
 *
 * INPUT:	u,v     (point of evaluation z = u+iv)
 *          p       (max. polynomial degree)
 *          fac     (factor to multiply)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *        / 1
 *        |
 *  fac * |  P_k(t) / (z-t)^2 dt;   z = u+iv, k = 0, ..., p
 *        |
 *        / -1
 *
 *   where P_k(t) are the Legendre polynomials.
 *
 */
void qt1(double u, double v, int const p, double const fac, 
         double *Vre, double *Vim){
  int j;
  double u2, v2, tmp, tmp1, x, y, absw2;    
  double um1, up1, logtmp, argtmp; 
  um1 = u-1.0; up1 = u+1.0;
  u2 = u*u;  
	v2 = v*v;
  
	phiinv(u,v, &x, &y);
	absw2 = x*x+y*y;

	if(absw2<6){
    /* Set V[0] and V[1] */
		if ( fabs(v) > eps ){ /* z is complex */
		  tmp1 = up1*um1-v2;
			tmp = 2.0/(tmp1*tmp1+4*u2*v2);
		 	Vre[0] =   fac * tmp1* tmp;
    	Vim[0] =  -fac *   2*u*v   * tmp;
    	if(p>0){
      	logtmp = get_logtmp(u, v, u2, v2, um1, up1); 
      	argtmp = get_argtmp(u, v, u2, v2, um1, up1); 
      	Vre[1] = u*Vre[0]-v*Vim[0] - fac*logtmp;
      	Vim[1] = u*Vim[0]+v*Vre[0] - fac*argtmp;
    	}
  	} else if(fabs(fabs(u)-1)>eps){ /* z is real and not +/-1 */   
    	Vre[0] = fac * 2.0/(up1*um1);    
			Vim[0] = 0.0;
    	if(p>0){
      	Vre[1] = u*Vre[0] - fac*log(fabs(up1/um1));
      	Vim[1] = 0.0;
    	} 
		} else { /* z = +/- 1 (Singularity --> Set values to 0)*/ 
    	for(j=0; j<=p; j++){
				Vre[j] = 0.0;
      	Vim[j] = 0.0;
    	}
    	return;
  	}

  }else{
	  double coeffs1[35]; /* maximal value of nu for |w|=6 */
  	int k,nu;
		if(p>0){
 			nu = get_taylor_degk(absw2);
 			coeffs1[0] = 0.;
 			for (k=1; k<nu+1; k++)  
   			coeffs1[k] = 16.0*k*(k+1.)/(2*k+1);
		}

		/* Set V[0] and V[1] */
		if ( fabs(v) > eps ){ /* z is complex */ 
		 	tmp1 = up1*um1-v2;
		 	tmp = 2.0/(tmp1*tmp1+4*u2*v2);
	 	 	Vre[0] =   fac * tmp1* tmp;
    	Vim[0] =  -fac * 2*u*v   * tmp;
     	if(p>0){
			  fcomplex ctmp;
      	tmp = absw2*absw2;
      	/* Horner scheme, x = 1/z^2 */
      	horner((x+y)*(x-y)/tmp, -2*x*y/tmp, coeffs1, nu, &ctmp.r, &ctmp.i);
      	ctmp = Cmul(ctmp,Complex(x/absw2,-y/absw2));
      	Vre[1] =fac*ctmp.r;
      	Vim[1] =fac*ctmp.i;             
    	}
    
   	} else{ /* z is real and not +/-1 */   
   		Vre[0] = fac * 2.0/(u2-1);
   		Vim[0] = 0.0;
   		if(p>0){ 
		  	horner(1./(x*x),0, coeffs1, nu, Vre+1, Vim+1);
     		Vre[1] = fac/x*Vre[1];
     		Vim[1] = 0.0;
    	} 
		} 
  }
  
	/* Calculate V[2]..V[p] */
  if(p>1){ 
    /* Inside ellipse, calculation via forward recurrene */
    if (checkEllipse(u2,v2,p)) { /*inside*/
      forward(1, 1.0, 1.0, 0.0, u, v, p, Vre, Vim);
    } else { /*outside*/
      backward (0, 1.0, -1.0, u, v, p, Vre, Vim);  
    }
  }    
}

/* qtn
 *
 * INPUT:	z       (point of evaluation)
 *          p       (max. polynomial degree)
 *          n       (max. exponent of (z-t)^n)
 *          fac     (factor to multiply)
 *          Vtmpre  (temporary variables)
 *          Vtmpim  (temporary variables)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)

 *        / 1
 *        | 
 *  fac * |  P_k(t) * log(t-z) dt;   z = u+iv, k = 0, ..., p 
 *        |                          
 *        / -1
 * and for j = 0..n
 *        / 1
 *        |
 *  fac * |  P_k(t) / (z-t)^j dt;   z = u+iv, k = 0, ..., p 
 *        |                          
 *        / -1
 *
 *   where P_k(t) are the Legendre polynomials.
 *
 */
void qtn(fcomplex const z, int const p, int const n, double const fac, 
         double *Vtmpre, double *Vtmpim, double *Vre, double *Vim){  
    int k,i;
    double u,u2,v2,ar,br;
    fcomplex z2, a,b, ctmp, ctmp1, ctmp2, ctmp3;
    
    /* Get real(z) and imag(z) */
    z2 = Cmul(z,z);
    u  = z.r;
    u2 = u*u;
    v2 = z.i * z.i;
    ctmp = Complex(z2.r-1.0,z2.i);
    
    /* Set qt0  V */
    qt0(z.r,z.i,p,fac,Vtmpre,Vtmpim);
    for (k=0;k<p+1;k++) {
        Vre[k] = Vtmpre[k];
        Vim[k] = Vtmpim[k];
    }    
    
    /* Set qt1  V */
    if (n>0) {
        qt1(z.r,z.i,p,fac,Vtmpre,Vtmpim);
        for (k=0;k<p+1;k++) {
            Vre[k+p+1] = Vtmpre[k];
            Vim[k+p+1] = Vtmpim[k];
        }   
    } 
    
    if (n>1) {
      /* Loop for calculation of qti upto degree p*/        
      for (k=0;k<p+1;k++) {        
        /* Loop for caluclation upto qtn */
        for (i=1;i<n;i++) {
          /* case: z is complex */
          if (v2 > eps) {
            a = Cdiv(RCmul(2*i,z),RCmul(i+1,ctmp));
            b = Cdiv(RCmul((k-i+1.0)*(k+i),Complex(1.0,0)),RCmul(i*(i+1.0),ctmp));
            ctmp1 = Cmul(a,Complex(Vre[k+i*(p+1)],Vim[k+i*(p+1)]));
            ctmp2 = Cmul(b,Complex(Vre[k+(i-1)*(p+1)],Vim[k+(i-1)*(p+1)]));
            ctmp3 = Cadd(ctmp1,ctmp2);
            Vre[k+(i+1)*(p+1)] = ctmp3.r;
            Vim[k+(i+1)*(p+1)] = ctmp3.i;      
          /* case: z is real and not +1 or -1*/    
          } else if (fabs(fabs(u)-1)>eps) {
            ar = 2.0*i*u / ((i+1)*(u2-1.0));
            br = (k-i+1.0)*(k+i) / (i*(i+1.0)*(u2-1.0));
            Vre[k+(i+1)*(p+1)] = ar*Vre[k+i*(p+1)] + br*Vre[k+(i-1)*(p+1)];                   
            Vim[k+(i+1)*(p+1)] = 0.0;  
          /* case: z = +1 or z = -1 */    
          } else {
            Vre[k+(i+1)*(p+1)] = 0.0;
            Vim[k+(i+1)*(p+1)] = 0.0;                          
          }
        }                            
      }    
    }            
}

/* q0
 *
 * INPUT:	u,v     (point of evaluation z = u+iv)
 *          p       (max. polynomial degree)
 *          fac     (factor to multiply)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *              / 1
 *              |
 *  fac * 0.5 * |  P_k(t) / (z-t) dt;   z = u+iv, k = 0, ..., p
 *              |
 *              / -1
 *
 *   where P_k(t) are the Legendre polynomials.
 *
 */

void q0(double u, double v, int const p, double const fac, double *Vre, double *Vim)
{
  qt0(u,v,p,0.5*fac,Vre,Vim);
}    

/* q1
 *
 * INPUT:	  z       (point of evaluation)
 *          p       (max. polynomial degree)
 *          fac     (factor to multiply)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *                             / 1
 *                             |
 *  fac * -0.5*(z^2-1)^(1/2) * | P_k(t)/(z-t)^2 dt;  z = u+iv, k = 0, ..., p
 *                             |
 *                             / -1
 *  or
 *                            / 1
 *                            |
 *  fac * 0.5*(1-x^2)^(1/2) * | P_k(t)/(x-t)^2 dt; |x| < 1 real,k = 0,...,p
 *                            |
 *                            / -1
 *
 *   where P_k(t) are the Legendre polynomials.
 *
 */

void q1(double u, double v, int const p, double const fac, 
        double *Vre, double *Vim)
{    
  double v2,u2;
  u2 = u*u;
  v2 = v*v;

	/* z is complex or not in [-1,1] */ 
 /* if (fabs(v) > eps || u2 > 1) {  */
  	int k, csgn;
    double tmp, utmp, vtmp, Rsqrt, Isqrt; 
		
    utmp = u2-v2-1.0;
  	vtmp = 2*u*v;
    
    
    tmp = sqrt(utmp*utmp+vtmp*vtmp);    
    csgn = (-vtmp>0 || (vtmp==0 && utmp>=0)) ? (1) : (-1);
    
    
    Rsqrt = sqrt((tmp+utmp)/2); 
    Isqrt = sqrt((tmp-utmp)/2);
		
    /* ctmp = Csqrt(ctmp); */

  	qt1(u,v,p,-0.5*fac,Vre,Vim);
		for (k=0;k<p+1;k++) {
      tmp = Vre[k];
      Vre[k] = Rsqrt*Vre[k]-Isqrt*Vim[k];
      Vim[k] = Rsqrt*Vim[k]+Isqrt*tmp;
    }
/*  }  
//  else  {
//  	qt1(u,v,p,0.5*sqrt(1.-u2)*fac,Vre,Vim);
//  }   */ 
}    

/* qn
 *
 * INPUT:	z       (point of evaluation)
 *          p       (max. polynomial degree)
 *          n       (upper limit of calculation of qn)
 *          fac     (factor to multiply)
 *          Vtmore  (temporary variables)
 *          Vtmpim  (temporary variables)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *                             / 1
 *                             |
 *  fac * -0.5*(z^2-1)^(1/2) * | P_k(t)/(z-t)^m dt;  z = u+iv, k = 0, ..., p
 *                             |
 *                             / -1
 *  for
 *                            / 1
 *                            |
 *  fac * 0.5*(1-x^2)^(1/2) * | P_k(t)/(x-t)^m dt; |x| < 1 real,k = 0,...,p
 *                            |
 *                            / -1
 *
 *  for m = 0..n and where P_k(t) are the Legendre polynomials.
 *
 */

void qn(fcomplex const z, int const p, int const n, double const fac, double *Vtmpre, double *Vtmpim, double *Vre, double *Vim){
    
    int k,i;
    double u,u2,v2,ar,br;
    fcomplex z2, a,b, ctmp, ctmp1, ctmp2, ctmp3;
    
    /* Get real(z) and imag(z) */
    z2 = Cmul(z,z);
    u  = z.r;
    u2 = u*u;
    v2 = z.i * z.i;

    /* Set gt0  V */
    q0(z.r,z.i,p,fac,Vtmpre,Vtmpim);
    for (k=0;k<p+1;k++) {
        Vre[k] = Vtmpre[k];
        Vim[k] = Vtmpim[k];
    }    
    
    /* Set gt1  V */
    if (n>0) {
        q1(z.r, z.i ,p,fac,Vtmpre,Vtmpim);
        for (k=0;k<p+1;k++) {
            Vre[k+p+1] = Vtmpre[k];
            Vim[k+p+1] = Vtmpim[k];
        }   
    } 
    
    if (n>1) {
      /* Loop for calculation of qi upto degree p*/        
      for (k=0;k<p+1;k++) {        
        /* Loop for caluclation upto qn */
        for (i=0;i<n-1;i++) {
          /* case: z is complex */
          if (fabs(z.i) > eps) {
            a = Cdiv(RCmul(-2.0*(i+1),z),ctmp);
            b = Complex((k-i)*(k+i+1.0),0.0);
            ctmp1 = Cmul(a,Complex(Vre[k+(i+1)*(p+1)],Vim[k+(i+1)*(p+1)]));
            ctmp2 = Cmul(b,Complex(Vre[k+i*(p+1)],Vim[k+i*(p+1)]));
            ctmp3 = Cadd(ctmp1,ctmp2);
            Vre[k+(i+2)*(p+1)] = ctmp3.r;
            Vim[k+(i+2)*(p+1)] = ctmp3.i;      
          /* case: z is real and not in [-1,1] */    
          } else if (u2 > 1) {
            ar = -2.0*(i+1)*u / sqrt(u2-1.0);
            br = (k+i+1.0)*(k-i);
            Vre[k+(i+2)*(p+1)] = ar*Vre[k+(i+1)*(p+1)] + br*Vre[k+i*(p+1)];                   
            Vim[k+(i+2)*(p+1)] = 0.0; 
          /* case: z is real and in (-1,1) */    
          } else if (u2 < 1) {
            ar = -2.0*(i+1)*u / sqrt(1.0-u2);
            br = (k+i+1.0)*(k-i);
            Vre[k+(i+2)*(p+1)] = ar*Vre[k+(i+1)*(p+1)] - br*Vre[k+i*(p+1)];                   
            Vim[k+(i+2)*(p+1)] = 0.0;  
          /* case: z = +1 or z = -1 */                        
          } else {
            Vre[k+(i+2)*(p+1)] = 0.0;
            Vim[k+(i+2)*(p+1)] = 0.0;                          
          }
        }                            
      }    
    }            
}


/* r0
 *
 * INPUT:		u,v     (point of evaluation z = u+iv)
 *          fac     (factor to multiply)
 *          p       (max. polynomial degree)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *         / 1
 *         |   N_k(t)
 *   fac * |  -------- dt;   z = u+iv, k = 1, ..., p
 *         |    z-t
 *         / -1
 *
 *                  / t
 *   where N_k(t) = |    P_{k-2}(s) ds
 *               -1 /
 *
 *
 */

void 
r0(double u, double v, int const p, double const fac,double *Vre, double *Vim){

	double tmp, x,y, absw2, u2, v2; 
	u2 = u*u; 
	v2 = v*v;
	phiinv(u,v,&x,&y);
	absw2 = x*x+y*y;

	if(absw2<2.25){
		double logtmp, argtmp, um1, up1, fac2; 
  	up1 = u+1;
  	um1 = u-1;

		if ( fabs(v) > eps ){ /* case: z is complex */
			logtmp = get_logtmp(u, v, u2, v2, um1, up1); 
      argtmp = get_argtmp(u, v, u2, v2, um1, up1); 
      fac2 = 0.5*fac;
      Vre[0] =   fac2 * (-um1*logtmp +   v*argtmp +2); 
      Vim[0] =  -fac2 * (   v*logtmp + um1*argtmp);  
      if(p>0){
        Vre[1] =  fac2 * (up1*logtmp - v*argtmp -2);
        Vim[1] =  fac2 * (up1*argtmp + v*logtmp);
        if(p>1){
          tmp = 0.5*(up1*um1-v2);  
          Vre[2] =  fac*(tmp*logtmp - u*v*argtmp - u);
          Vim[2] =  fac*(tmp*argtmp + u*v*logtmp - v);
          if(p>2){
            Vre[3] = u*Vre[2] - v*Vim[2] + 2.0/3.0*fac;
            Vim[3] = u*Vim[2] + v*Vre[2];   
          }
        }
      }
    } else if (fabs(fabs(u)-1)>eps){ /* case: z is real and z != +/- 1 */   
			logtmp = 0.5*log(fabs(up1/um1));      
      Vre[0] = fac*(logtmp*(1.0-u)+1.0);
      Vim[0] = 0.0; 
      if(p>0){
        Vre[1] =  fac*(logtmp*up1-1.0);
        Vim[1] =  0.0;
        if(p>1){
          Vre[2] =  fac*(up1*um1*logtmp-u);
          Vim[2] =  0.0;
          if(p>2){
            Vre[3] = u*Vre[2]  + 2.0/3.0*fac;
            Vim[3] = 0.0;
          }
        }
      }  
		} else { /* case: z = 1 or z = -1 */  
      Vre[0] =  fac;  Vim[0] =  0.0;
      if(p>0){
        Vre[1] =  -fac; Vim[1] =  0.0;
        if(p>1){
          Vre[2] = (u>0)?(-fac):(fac); Vim[2] = 0.0;  
          if(p>2){
            Vre[3] = -fac/3.0; Vim[3] = 0.0;   
          }
        }
      }
		}

	}else{
    double coeffs0[120], coeffs1[120], coeffs2[40], coeffs3[40]; 
  	int k,nu0, nu1;
		/* maximal value of nu for |w|=1.5 */
 		nu0 = get_taylor_degk(absw2);
 		nu1 = get_taylor_degik3(absw2);
    coeffs0[0] = 0.;	
		coeffs1[0] = 0.;
    for (k=1; k<nu0+1; k++){  
      coeffs0[2*k]   = -4.*k/((2.*k-1.)*(2.*k+1.));
      coeffs0[2*k-1] =  2.  / (2.*k-1.);
			coeffs1[2*k]   = -coeffs0[2*k]; 
			coeffs1[2*k-1] =  coeffs0[2*k-1];
		}
		if(p>1){
			for(k=0; k<nu1+1; k++)	
				coeffs2[k] = 4./((2.*k-1.)*(2.*k+1.)*(2.*k+3.));
			if(p>2){
				coeffs3[0] = 0.;
				for(k=1; k<=nu1; k++)	
					coeffs3[k] = 8.*k/((2.*k-3.)*(2.*k-1.)*(2.*k+1.)*(2.*k+3.));
			}
		}

		if(fabs(v)>eps){
			/* Horner scheme, 1/w */
    	horner(x/absw2, -y/absw2, coeffs0, 2*nu0, Vre, Vim);
			Vre[0] *= fac;
			Vim[0] *= fac;
			if(p>0){
				/* Horner scheme,1/w */
    		horner(x/absw2, -y/absw2, coeffs1, 2*nu0, Vre+1, Vim+1);
				Vre[1] *= fac;
				Vim[1] *= fac;
				if(p>1){
					fcomplex ctmp;
 	     		/* Horner scheme, 1/w^2 */
					tmp = absw2*absw2;
      		horner((x+y)*(x-y)/tmp, -2*x*y/tmp, coeffs2, nu1, &ctmp.r, &ctmp.i);
      		ctmp = Cmul(ctmp,Complex(x/absw2,-y/absw2));
					Vre[2] = fac*ctmp.r;
					Vim[2] = fac*ctmp.i;
					if(p>2){
      			horner((x+y)*(x-y)/tmp, -2*x*y/tmp, coeffs3, nu1, &ctmp.r, &ctmp.i);
						Vre[3] = fac*ctmp.r;
						Vim[3] = fac*ctmp.i;
					}
				}
			}
		}else{
		 tmp = 1.0/x;
		/* Horner scheme, 1/x */
    	horner(tmp,0., coeffs0, 2*nu0, Vre, Vim);
			Vre[0] *= fac;
			Vim[0]  = 0.;
    
			if(p>0){
				/* Horner scheme,1/x */
    		horner(tmp, 0., coeffs1, 2*nu0, Vre+1, Vim+1);
				Vre[1] *= fac;
				Vim[1]  = 0.;
				if(p>1){
 	   			/* Horner scheme,  1/x^2 */
      		horner(tmp*tmp, 0., coeffs2, nu1, Vre+2, Vim+2);
					Vre[2] = fac*tmp*Vre[2];
					Vim[2] = 0.;
					if(p>2){
      			horner(tmp*tmp, 0., coeffs3, nu1, Vre+3, Vim+3);
						Vre[3] *= fac;
						Vim[3]  = 0.;
					}
				}
			}
			
		}
	}

  /* Calculate the remaining values of V */
  if(p>3){ 
    if (checkEllipse(u2,v2,p)) { /*inside*/
      forward(3, -1.0, -2.0, 1.0, u, v, p, Vre, Vim); 		 
    } else { /*outside*/
      backward (1, 1.0, 1.0, u, v, p, Vre, Vim);   
    }
  }  
}

/* r1
 *
 * INPUT:		u,v     (point of evaluation z = u+iv)
 *          fac     (factor to multiply)
 *          p       (max. polynomial degree)
 * OUTPUT:  Vre     (real part of the integral)
 *          Vim     (imaginary part of the integral)
 *
 *         / 1
 *         |     N_k(t)
 *   fac * |  ----------- dt;   z = u+iv, k = 1, ..., p
 *         |    (z-t)^2
 *         / -1
 *
 *                  / t
 *   where N_j(t) = |    P_{k-2}(s) ds 
 *               -1 /
 *
 *
 */

void r1(double u, double v, int const p, double const fac, 
        double *Vre, double *Vim){

	double tmp, x, y, absw2, u2, v2; 

	u2 = u*u;
	v2 = v*v;
	phiinv(u,v,&x,&y);
	absw2 = x*x+y*y;
      		
  /* Set initial values V[0], V[1], V[2] and V[3]*/
	if(absw2<2.25){
		double up1, um1, logtmp, argtmp;
  	up1 = u+1.0;
  	um1 = u-1.0;
		
		if ( fabs(v) > eps ){ /* case: z is complex */
			logtmp = get_logtmp(u, v, u2, v2, um1, up1); 
      argtmp = get_argtmp(u, v, u2, v2, um1, up1); 
      tmp = 1./(up1*up1+v2);
      Vre[0] = fac*(0.5*logtmp - up1*tmp);
      Vim[0] = fac*(  v*tmp    + 0.5*argtmp);
      if(p>0){
      	tmp = 1./(um1*um1+v2);
      	Vre[1] = fac*(-0.5*logtmp + um1*tmp);
      	Vim[1] =-fac*(   v*tmp    + 0.5*argtmp);
      	if(p>1){
        	Vre[2] = fac*(-u*logtmp+v*argtmp+2);
        	Vim[2] =-fac*( v*logtmp+u*argtmp);
        	if(p>2){
          	tmp = 0.5-1.5*(u2-v2);
          	Vre[3] = fac*(tmp*logtmp + 3.*u*(v*argtmp+1.));
          	Vim[3] = fac*(tmp*argtmp - 3.*v*(u*logtmp-1.));
        	}
      	}
			}
		} else if (fabs(fabs(u)-1)>eps){  /* case: z is real, z != +/- 1 */ 
      logtmp = log(fabs(up1/um1));
      Vre[0] = fac*(-1./up1+0.5*logtmp);
      Vim[0] = 0.0;
			if(p>1){
      	Vre[1] = fac*(1./um1 - 0.5*logtmp);
      	Vim[1] = 0.0;
      	if(p>1){
        	Vre[2] = fac*(-u*logtmp+2.0);
        	Vim[2] = 0.0;  
        	if(p>2){
          	Vre[3] = fac*((0.5-1.5*u2)*logtmp+3.*u);
          	Vim[3] = 0.0;          
        	}
      	}
			}
    } else if(u>0){    /* case: z = + 1 --> Cauchy Principal value*/ 
      Vre[0] = -fac*0.5; Vim[0] = 0.0;
			if(p>0){
      	Vre[1] = 0.0; Vim[1] = 0.0;        
      	if(p>1){
        	Vre[2] = fac*2.0; Vim[2] = 0.0;            
      		if(p>2){
        		Vre[3] = fac*3.0;	Vim[3] = 0.0; 
     			}
				}
			}
    } else {            /* case: z = - 1 --> Cauchy Principal value*/ 
      Vre[0] = 0.0; Vim[0] = 0.0;
			if(p>0){
      	Vre[1] = fac*0.5; Vim[1] = 0.0;        
      	if(p>1){
        	Vre[2] = fac*2.0;	Vim[2] = 0.0;            
      		if(p>2){
        		Vre[3] = -fac*3.0; Vim[3] = 0.0; 
					}
				}
      }   
    }
	}else{
 		double coeffs0[110], coeffs1[110], coeffs2[55], coeffs3[55]; 
		/* maximal value of nu for |w|=1.5 */
  	int k,nu;
 		nu = get_taylor_degik(absw2);

		coeffs0[0] = 0.; coeffs0[1] = 0.;
		coeffs1[0] = 0.; coeffs1[1] = 0.;
		for(k=1; k<=nu; k++){
			coeffs0[2*k]   = 4*k;
			coeffs0[2*k+1] = -8*k*(k+1.)/(2.*k+1.);
			coeffs1[2*k]   = coeffs0[2*k];
			coeffs1[2*k+1] = -coeffs0[2*k+1];
		}

		if(p>1){
			coeffs2[0] = 0.;
			for(k=1; k<=nu; k++)	coeffs2[k] = -8.*k/((2.*k-1.)*(2*k+1));
			if(p>2){
				coeffs3[0] = 0.;
				for(k=1; k<=nu; k++){
					coeffs3[k] = -16.*k*(k+1.)/((2.*k-1.)*(2.*k+1.)*(2.*k+3.));
				}
			}
		}

		if(fabs(v)>eps){
			/* Horner scheme, x = 1/z */
    	horner(x/absw2, -y/absw2, coeffs0, 2*nu, Vre, Vim);
			Vre[0] *= fac;	Vim[0] *= fac;
    
			if(p>0){
				/* Horner scheme, x = 1/z */
    		horner(x/absw2, -y/absw2, coeffs1, 2*nu, Vre+1, Vim+1);
				Vre[1] *= fac;	Vim[1] *= fac;
				if(p>1){
 	     		/* Horner scheme, x = 1/z^2 */
					tmp = absw2*absw2;
      		horner((x+y)*(x-y)/tmp, -2*x*y/tmp, coeffs2, nu, Vre+2, Vim+2);
					Vre[2] *= fac;	Vim[2] *= fac;
					if(p>2){
						fcomplex ctmp;
      			horner((x+y)*(x-y)/tmp, -2*x*y/tmp, coeffs3, nu, &ctmp.r, &ctmp.i);
						ctmp = Cmul(ctmp, Complex(x/absw2, -y/absw2));
						Vre[3] = fac*ctmp.r;
						Vim[3] = fac*ctmp.i;
					}
				}
			}
		}else{
			tmp = 1.0/x;
			/* Horner scheme, x = 1/z */
    	horner(tmp,0., coeffs0, 2*nu, Vre, Vim);
			Vre[0] *= fac;
			Vim[0]  = 0.;
    
			if(p>0){
				/* Horner scheme, x = 1/z */
    		horner(tmp, 0., coeffs1, 2*nu, Vre+1, Vim+1);
				Vre[1] *= fac;
				Vim[1]  = 0.;
				if(p>1){
 	    		/* Horner scheme, x = 1/z^2 */
      		horner(tmp*tmp, 0., coeffs2, nu, Vre+2, Vim+2);
					Vre[2] *= fac;
					Vim[2] = 0.;
					if(p>2){
      			horner(tmp*tmp, 0., coeffs3, nu, Vre+3, Vim+3);
						Vre[3] = fac*tmp*Vre[3];
						Vim[3] = 0.;
					}
				}
			}
		}
	}

  /* Calculate the remaining values of V */
  if(p>3){ 
    if (checkEllipse(u2,v2,p)) {  /*inside*/
      forward(3, -1.0, -1.0, 0.0, u, v, p, Vre, Vim);
    } else { /*outside*/
      backward(1 , 0.0, 0.0, u, v, p, Vre, Vim); 
    }
  }
}




void backward_old (int swt, int end, int end1, double a, double b, \
                      double u, double v, int p, double *Vre, double *Vim){
  double rretmp, rimtmp;
  double tmp1, tmp2,tmp3;
  int n, k, k1, k2; 
  int nu;
  double loginvtol =  3.684136148790473e+01;  /* log(10^16) */
	fcomplex z2, tmp;
    
  /* Assign nu*/
  z2 = Complex(fabs(u),fabs(v));
  tmp = Csqrt(Cadd(Complex(-1.,0.), Cmul(z2,z2)));	      
  nu = (int) (p+ceil(loginvtol / (2.*log(Cabs(Cadd(z2, tmp)))))); 

  /* Calculate vector r */
  rretmp = 1.0; rimtmp = 0.0;
  for (k=nu; k>=p-swt+2; k--){   
    tmp1 = u-rretmp;
    tmp2 = v-rimtmp;
    tmp3 = (k*k-a)/((4*k*k-1.0)*(tmp1*tmp1 + tmp2*tmp2));
    rretmp =  tmp3*tmp1;
    rimtmp = -tmp3*tmp2;  
  }
  
  Vre[p] = rretmp;
  Vim[p] = rimtmp;  
  for (n=p; n>=end+swt; n--){   
    k = n-swt+1;
    tmp1 = u-Vre[n];
    tmp2 = v-Vim[n];
    tmp3 = (k*k-a)/((4*k*k-1.0)*(tmp1*tmp1 + tmp2*tmp2));
    Vre[n-1] =  tmp3*tmp1;
    Vim[n-1] = -tmp3*tmp2;    
  }  

  /* Calculate remaining values of V */
  k1 = swt-2;
  k2 = swt-3;
  for (n=swt; n<p+end1; n++){
    tmp1 = (2*n-1.0)/(n+b);
    tmp2 = Vre[n+k1];
    Vre[n+k1] = tmp1*(Vre[n+k1]*Vre[n+k2]-Vim[n+k1]*Vim[n+k2]);
    Vim[n+k1] = tmp1*(tmp2*Vim[n+k2]+Vim[n+k1]*Vre[n+k2]);
  }
}

