% buildSymmRHS_f  Compute the right hand side of Symm's integral
%    equation
%
%    b = buildSymmRHS_f(coordinates,elements,f,p[,options])
%    computes a vector b which contains the L2-scalar products on 
%    \Gamma of a function f given by a funciton pointer with all 
%    basis functions of S0 using Gauss quadrature. 
%    \Gamma is a polygonial boundary described by coordinats and elements 
%    and p contains the polynomial degrees of S0 on each element.
%
%  Example:
%    coordinates = [-1,0; 1,0; 1,1];
%    elements    = [1,2;2,3];
%    p = [2,1];
%    f = @(x,varargin) [sin(pi*x(:,1)),x(:,2)];  
%    b = buildSymmRHS_f(coordinates,elements,f,p)
%
%    b =
% 
%        -0.0000         0
%         0.6366         0
%         0.0000         0
%         0.0000    0.5000
%        -0.0000    0.1667

function b = buildSymmRHS_f(coordinates,elements,f,p)

p = checkp0(p, size(elements,1));

ngauss = 8;
while ngauss <= max(p)+8
  ngauss = 2*ngauss;
end
ngauss=2*ngauss;
ngauss = min(ngauss, 8192);

%*** Get Gauss weights and nodes and evaluate Legendre Polynomials at nodes
[weights,xg] = gauss(ngauss);
pleg = bLegendre(xg,max(p));

%*** Calculate u and m for each element and claculate mesh width
%TO DO

cp = cumsum([0;p(:)+1]);
b  = zeros(cp(end),1);  
for k = 1:size(elements,1)
  % DO THE QUADRATURE - IF POSSIBLE VECTORIZED
  
  tmp = 0;
  index = cp(k)+1:cp(k+1);
  b(index,:) = tmp; 
end



% checkp0 Ensure that p0 has the right size.
% 
%    p0 = checkp0(p0,N) 
%    Ensures that p0 is a vector of size (N,1) containing only non-negative
%    values. 
%    If the input argument p0 is a scalar, the output is a vector of size
%    (N,1) where all values are that scalar.
function p0 = checkp0(p0,N)

if length(p0) == 1 && p0>=0
  p0 = p0*ones(N,1);
elseif numel(p0) == N && min(p0)>=0
  p0 = p0(:);
else
  error('p0 must be a vector of length size(elements,1) or a non-negative integer.')
end
