% plotArclengthP   plot a function in P \subset H^{-1/2}(\Gamma) defined by 
%                  its coefficients with respect to the Legendre Polynomials
%                  over the arclenth of \Gamma. 
%
%   plotArclengthP(coordinates,elements,p,x,varargin)
%   draws the solution given by the coefficient vector x for a boundary 
%   \Gamma which is specified by coordinates and elements. p (either a vector 
%   of length size(elements,1) of an integer >=0 species the polynomial degrees 
%   for the ansatz space P. 
%   
%
%  Examples: 
%    elements    = [1,2; 2,3; 3,4];
%    coordinates = [0,0; 1,0; 1,1; 0,1];
%    x = [1,1,2,1,2,3].';
%    p = [0,1,2];
%    addpath('../laplace/mat') 
%    addpath('./c')
%    plotArclengthP(coordinates,elements,p,x, @(x,n,options) x(:,1));

function plotArclengthP(coordinates,elements,p,x,varargin)

%*** check the optional input arguments
if nargin < 4
  error('At least four input arguments required.')
else
  options = [];
  fig = 0;
  phi = false;
  for k = 1:nargin - 4
    if isstruct(varargin{k})
      options = varargin{k};
    elseif isa(varargin{k}, 'function_handle')
      phi = varargin{k};
    elseif isscalar(varargin{k}) && isnumeric(varargin{k})
      fig = varargin{k};
    end
  end
end
if fig > 0
    figure(fig);
 elseif fig == 0
    cla(gca);
end

p = checkp0(p,size(elements,1));
nC = length(unique(elements)); % number of coordinates
nE = size(elements,1);         % number of elements
dim = size(x,2);               % problem dimension

%*** plot some points for the legend
legstr = 'leg = legend(';
if dim==1
  plot(0,0); hold on
else
  plot(0,0,'b', 0,0,'r'); hold on
end
if isa(phi,'function_handle')
  if dim ==1 
    plot(0,0,'--c')
  else
    plot(0,0,'--c',0,0,'--m')
  end  
end

%*** use first node of triangulation
node = coordinates(elements(1,1),:);

%*** determine neighbour Ek of Ej with respect to second node of Ej
node2element = zeros(nC,1);
node2element(elements(:,1)) = (1:nE)';
element2neighbour = node2element(elements(:,2));

%*** determine element E0 which with initial node closest to prescribed node
[~,initial] = min(sum( (coordinates(elements(:,1),:)-repmat(node,nE,1)).^2, 2));
initial = initial(1);

%*** if Gamma is not connected, we restrict to one connectedness component
idx = ones(nE,1)*initial;
for j = 2:nE
    idx(j) = element2neighbour(idx(j-1));
    if idx(j) == initial;
        break
    end
end

if(isempty(j))
  nE = 1;
else
  nE=j;
end
%*** re-order elements such that E1 = E0 und that Ej is neighbor of Ej-1
elements = elements(idx(1:nE),:);
ps = p(idx(1:nE));

cp  = cumsum([0;p]+1);
cps = cumsum([0;ps]+1);
xs = zeros(size(x));
for k = 1:nE
  xs((cps(k):cps(k+1)-1),:) = x((cp(idx(k)):cp(idx(k)+1)-1),:);
end


%*** compute local mesh-size
h = sqrt(sum((coordinates(elements(:,1),:)-coordinates(elements(:,2),:)).^2,2));

%*** compute arclength s.t. [s(j),s(j+1)] corresponds to Ej
s = [0;cumsum(h)];

%*** evaluate Legendre polynomials
pts = linspace(-0.95,0.95,5*(max(p)+1));
pleg = bLegendre(pts,max(p));


%*** plot discrete solution over arclength
for i = 1:nE
  phih = pleg(:,1:ps(i)+1)*xs(cps(i):cps(i+1)-1,:);
  xvals = (s(i+1)+s(i))/2 + pts*(s(i+1)-s(i))/2;
  
  plot(xvals,phih(:,1))  
  hold on
  if dim ==2
     plot(xvals,phih(:,2),'r')
  end
end

%* for the legend
if dim==1
  legstr = [legstr,'''\phi_h'''];
else
  legstr = [legstr,'''\phi_{h,1}'',''\phi_{h,2}'''];
end

%*** optionally, we plot the exact solution
if isa(phi,'function_handle')
    %*** evaluation points for phi per element w.r.t. reference element [0,1]
    evalpts = linspace(0.1,0.9,200);    
    for i = 1:nE
      %*** compute all evaluation points for phi
      a = coordinates(elements(i,1),:);
      b = coordinates(elements(i,2),:);
    
      n = b-a;
      n = [n(2),-n(1)]./norm(n);
    
      sx = (1-evalpts)'*a + evalpts'*b;
      phi_sx = phi(sx,n,options);
      s_sx = s(i)+evalpts'*h(i);

      plot(s_sx,phi_sx(:,1),'c--')      
      if dim ==2
        plot(s_sx,phi_sx(:,2),'m--')
      end 
    end
   
    %* for the legend
    if dim ==1
      legstr = [legstr,',''\phi'''];
    else
      legstr = [legstr,',''\phi_1'',''\phi_2'''];
    end
end
hold off
eval([legstr,');'])
set(leg,'location', 'EastOutside')